@@ -3,9 +3,10 @@
 from os.path import join as pathjoin
 from typing import TypeVar
 
-from ..loader import load_problem, load_symbol_from_dir
+from ..loader import load_problem
+from ..runner import load_and_run
 from .metadata import load_submission_metadata_from_path
-from .runner import run_suite
+from .schema import write_to
 
 Output = TypeVar("Output")
 
@@ -27,14 +28,8 @@ def gradescope_main(
     The parameters are to allow overriding the file structure in unit testing.
     """
     problem = load_problem(problem_src)  # type: ignore
-    under_test = load_symbol_from_dir(
-        submission_dir, problem.expected_symbol()  # pylint: disable=no-member
-    )
     metadata = load_submission_metadata_from_path(metadata_file)
-
-    suite = problem.generate_test_suite(
-        under_test, metadata.assignment.total_points
-    )  # pylint: disable=no-member
+    output = load_and_run(problem, submission_dir, metadata.assignment.total_points)
 
     with open(output_file, "w+", encoding="UTF-8") as file:
-        run_suite(suite, file)
+        write_to(output, file)
